import json
import logging
from typing import Optional
from unittest import TestCase, mock

import pytest

from data_config import read_config_file
from keepercommander.params import KeeperParams
from keepercommander import cli, api
from keepercommander.commands import aram, enterprise


@pytest.mark.integration
class TestEnterpriseCommands(TestCase):
    params = None   # type: Optional[KeeperParams]

    @classmethod
    def setUpClass(cls):
        cls.params = KeeperParams()
        read_config_file(cls.params, 'enterprise.json')
        api.login(cls.params)
        TestEnterpriseCommands.wipe_out_data()

    @classmethod
    def tearDownClass(cls):
        params = cls.params
        api.query_enterprise(params)
        for user in params.enterprise['users']:
            if user['status'] == 'actove' and user['lock'] != 0:
                request = {
                    'command': 'enterprise_user_lock',
                    'enterprise_user_id': user['enterprise_user_id'],
                    'lock': 'unlocked'
                }
                api.communicate(params, request)

        cli.do_command(params, 'logout')

    @classmethod
    def wipe_out_data(cls):
        params = cls.params    # type: KeeperParams
        managed_roles = set()
        if 'managed_nodes' in params.enterprise:
            for mn in params.enterprise['managed_nodes']:
                managed_roles.add(mn['role_id'])

        if 'role_users' in params.enterprise:
            for ru in params.enterprise['role_users']:
                if ru['role_id'] not in managed_roles:
                    request = {
                        'command': 'role_user_remove',
                        'role_id': ru['role_id'],
                        'enterprise_user_id': ru['enterprise_user_id']
                    }
                    api.communicate(params, request)

        if 'users' in params.enterprise:
            for user in params.enterprise['users']:
                if user['status'] == 'invited':
                    request = {
                        'command': 'enterprise_user_delete',
                        'enterprise_user_id': user['enterprise_user_id']
                    }
                    api.communicate(params, request)

        if 'teams' in params.enterprise:
            for team in params.enterprise['teams']:
                request = {
                    'command': 'team_delete',
                    'team_uid': team['team_uid']
                }
                api.communicate(params, request)
        api.query_enterprise(params)

    def test_commands(self):
        params = TestEnterpriseCommands.params    # type: KeeperParams
        self.assertIsNotNone(params.enterprise)
        test_user = params.config['user']
        new_user = 'integration.user@keepersecurity.com'
        new_team = 'Test Team'
        with mock.patch('builtins.input', side_effect=KeyboardInterrupt()), mock.patch('builtins.print'):
            cli.do_command(params, 'enterprise-info --verbose')

            cli.do_command(params, 'enterprise-team --add "{0}"'.format(new_team))
            cli.do_command(params, 'enterprise-team --restrict-edit=on --restrict-share=on --restrict-view=off "{0}"'.format(new_team))

            cli.do_command(params, 'enterprise-user --add --name="Test User" {0}'.format(new_user))
            cli.do_command(params, 'enterprise-user --lock "{0}"'.format(test_user))
            cli.do_command(params, 'enterprise-user --unlock "{0}"'.format(test_user))

            cli.do_command(params, 'enterprise-team --add-user="{0}" "{1}"'.format(test_user, new_team))
            cli.do_command(params, 'enterprise-team --remove-user="{0}" "{1}"'.format(test_user, new_team))
            cli.do_command(params, 'enterprise-user --add-team="{0}" "{1}"'.format(new_team, test_user))
            cli.do_command(params, 'enterprise-user --remove-team="{0}" "{1}"'.format(new_team, test_user))

            role_id = None
            managed_roles = set()
            for mn in params.enterprise['managed_nodes']:
                managed_roles.add(mn['role_id'])

            for role in params.enterprise['roles']:
                if role['role_id'] not in managed_roles:
                    role_id = role['role_id']
                    break
            if role_id:
                cli.do_command(params, 'enterprise-role --add-user="{0}" "{1}"'.format(new_user, role_id))

    def test_add_enterprise_user(self):
        params = TestEnterpriseCommands.params    # type: KeeperParams
        self.assertIsNotNone(params.enterprise)

        template_body = '''
[
    {
        "title": "Record For ${user_name}",
        "login": "${user_email}",
        "password": "${generate_password}",
        "login_url": "https://keepersecurity.com",
        "notes": "notes",
        "custom_fields": {
            "key1": "value1",
            "key2": "value2"
        }
    },
    {
        "title": "Empty record"
    }

]'''

        new_user = 'integration.new.user@keepersecurity.com'
        with mock.patch('builtins.open', mock.mock_open(read_data=template_body)), \
                mock.patch('os.path.abspath', return_value='template.json'), \
                mock.patch('os.path.isfile', return_value=True):

            with self.assertLogs(level=logging.WARNING):
                cli.do_command(params, f'create-user --generate --name="New User" --expire --records="template.json" '
                                       f'--question="This app name?" --answer="Commander" {new_user}')

    def test_report_commands(self):
        params = TestEnterpriseCommands.params
        self.assertIsNotNone(params.enterprise)

        user_report = enterprise.UserReportCommand()
        report_json = user_report.execute(params,  format='json')
        report = json.loads(report_json)
        self.assertTrue(isinstance(report, list))

        device_approve = enterprise.DeviceApproveCommand()
        device_approve.execute(params, reload=True, format='json')

        enterprise_info = enterprise.EnterpriseInfoCommand()
        report_csv = enterprise_info.execute(params, nodes=True, columns=','.join(enterprise.SUPPORTED_NODE_COLUMNS), format='csv')
        self.assertTrue(len(report_csv) > 0)
        report_json = enterprise_info.execute(params, users=True, columns=','.join(enterprise.SUPPORTED_USER_COLUMNS), format='json')
        report = json.loads(report_json)
        self.assertTrue(isinstance(report, list))
        report_json = enterprise_info.execute(params, teams=True, columns=','.join(enterprise.SUPPORTED_TEAM_COLUMNS), format='json')
        report = json.loads(report_json)
        self.assertTrue(isinstance(report, list))
        report_csv = enterprise_info.execute(params, roles=True, columns=','.join(enterprise.SUPPORTED_ROLE_COLUMNS), format='csv')
        self.assertTrue(len(report_csv) > 0)

        audit_report = aram.AuditReportCommand()
        report_json = audit_report.execute(params, report_type='dim', columns=['audit_event_type'], format='json')
        report = json.loads(report_json)
        self.assertTrue(isinstance(report, list))
        report_json = audit_report.execute(params, report_type='raw', created='last_30_days', format='json')
        report = json.loads(report_json)
        self.assertTrue(isinstance(report, list))
        report_json = audit_report.execute(params, report_type='day', columns=['audit_event_type'], aggregate=['occurrences', 'first_created', 'last_created'],
                                           created='last_30_days', format='json')
        report = json.loads(report_json)
        self.assertTrue(isinstance(report, list))

        security_audit_report = enterprise.SecurityAuditReportCommand()
        report_json = security_audit_report.execute(params, format='json')
        report = json.loads(report_json)
        self.assertTrue(isinstance(report, list))

        report_json = security_audit_report.execute(params, breachwatch=True, format='json')
        report = json.loads(report_json)
        self.assertTrue(isinstance(report, list))



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2022 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import abc
import json
import logging
from typing import Optional, List, Set, Tuple, Dict

from google.protobuf import message

from .params import KeeperParams
from .proto import enterprise_pb2 as proto
from . import api, utils, crypto


def query_enterprise(params):  # type: (KeeperParams) -> None
    if not params.enterprise_loader:
        params.enterprise_loader = _EnterpriseLoader()
    params.enterprise_loader.load(params)


def _to_key_type(key_type):  # type: (proto.EncryptedKeyType) -> str
    if key_type == proto.KT_ENCRYPTED_BY_DATA_KEY:
        return 'encrypted_by_data_key'
    if key_type == proto.KT_ENCRYPTED_BY_PUBLIC_KEY:
        return 'encrypted_by_public_key'
    if key_type == proto.KT_ENCRYPTED_BY_DATA_KEY_GCM:
        return 'encrypted_by_data_key_gcm'
    if key_type == proto.KT_ENCRYPTED_BY_PUBLIC_KEY_ECC:
        return 'encrypted_by_public_key_ecc'
    return 'no_key'


class EnterpriseInfo(object):
    def __init__(self):
        self._tree_key = b''
        self._rsa_key = b''
        self._ec_key = b''
        self._enterprise_name = ''

    @property
    def tree_key(self):
        return self._tree_key

    @property
    def rsa_key(self):
        return self._rsa_key

    @property
    def ec_key(self):
        return self._ec_key

    @property
    def enterprise_name(self):
        return self._enterprise_name


class _EnterpriseLoader(object):
    def __init__(self):
        super(_EnterpriseLoader, self).__init__()
        self._enterprise = EnterpriseInfo()
        self._continuationToken = b''
        self._data_types = {   # type: dict[int, _EnterpriseDataParser]
            proto.NODES: _EnterpriseNodeEntity(self._enterprise),
            proto.USERS: _EnterpriseUserEntity(self._enterprise),
            proto.TEAMS: _EnterpriseTeamEntity(self._enterprise),
            proto.ROLES: _EnterpriseRoleEntity(self._enterprise),
            proto.LICENSES: _EnterpriseLicenseEntity(self._enterprise),
            proto.QUEUED_TEAMS: _EnterpriseQueuedTeamEntity(self._enterprise),
            proto.SCIMS: _EnterpriseScimEntity(self._enterprise),
            proto.SSO_SERVICES: _EnterpriseSsoServiceEntity(self._enterprise),
            proto.BRIDGES: _EnterpriseBridgeEntity(self._enterprise),
            proto.EMAIL_PROVISION: _EnterpriseEmailProvisionEntity(self._enterprise),
            proto.TEAM_USERS: _EnterpriseTeamUserEntity(self._enterprise),
            proto.QUEUED_TEAM_USERS: _EnterpriseQueuedTeamUserEntity(self._enterprise),
            proto.ROLE_USERS: _EnterpriseRoleUserEntity(self._enterprise),
            proto.ROLE_TEAMS: _EnterpriseRoleTeamEntity(self._enterprise),
            proto.MANAGED_NODES: _EnterpriseManagedNodeEntity(self._enterprise),
            proto.ROLE_PRIVILEGES: _EnterpriseRolePrivilegeEntity(self._enterprise),
            proto.ROLE_ENFORCEMENTS: _EnterpriseRoleEnforcements(self._enterprise),
            proto.MANAGED_COMPANIES: _EnterpriseManagedCompanyEntity(self._enterprise),
            proto.DEVICES_REQUEST_FOR_ADMIN_APPROVAL: _EnterpriseAdminApprovalRequestEntity(self._enterprise),
        }
        teams = self._data_types[proto.TEAMS]
        if isinstance(teams, _EnterpriseEntity):
            teams.register_link('team_uid', self._data_types[proto.TEAM_USERS])
            teams.register_link('team_uid', self._data_types[proto.ROLE_TEAMS])

        users = self._data_types[proto.USERS]
        if isinstance(teams, _EnterpriseEntity):
            users.register_link('enterprise_user_id', self._data_types[proto.TEAM_USERS])
            users.register_link('enterprise_user_id', self._data_types[proto.ROLE_USERS])

        roles = self._data_types[proto.ROLES]
        if isinstance(roles, _EnterpriseEntity):
            users.register_link('role_id', self._data_types[proto.ROLE_TEAMS])
            users.register_link('role_id', self._data_types[proto.ROLE_USERS])
            users.register_link('role_id', self._data_types[proto.MANAGED_NODES])

    @property
    def enterprise(self):
        return self._enterprise

    def load(self, params):  # type: (KeeperParams) -> None
        if params.enterprise is None:
            params.enterprise = {}
            self._continuationToken = b''

        if not self._enterprise.tree_key:
            rq = proto.GetEnterpriseDataKeysRequest()
            rs = api.communicate_rest(params, rq, 'enterprise/get_enterprise_data_keys',
                                      rs_type=proto.GetEnterpriseDataKeysResponse)
            if rs.treeKey:
                encrypted_tree_key = utils.base64_url_decode(rs.treeKey.treeKey)
                if rs.treeKey.keyTypeId == proto.ENCRYPTED_BY_DATA_KEY:
                    self._enterprise._tree_key = crypto.decrypt_aes_v1(encrypted_tree_key, params.data_key)
                elif rs.treeKey.keyTypeId == proto.ENCRYPTED_BY_PUBLIC_KEY:
                    if len(encrypted_tree_key) == 60:
                        self._enterprise._tree_key = crypto.decrypt_aes_v2(encrypted_tree_key, params.msp_tree_key)
                    else:
                        self._enterprise._tree_key = api.decrypt_rsa(rs.treeKey.treeKey, params.rsa_key)
                params.enterprise['unencrypted_tree_key'] = self._enterprise.tree_key

            keys = {}    # type: Dict[str, str]
            if rs.enterpriseKeys:
                if rs.enterpriseKeys.rsaEncryptedPrivateKey:
                    self._enterprise._rsa_key = \
                        api.decrypt_aes_plain(rs.enterpriseKeys.rsaEncryptedPrivateKey, self._enterprise.tree_key)
                    keys['rsa_public_key'] = utils.base64_url_encode(rs.enterpriseKeys.rsaPublicKey)
                    keys['rsa_encrypted_private_key'] = \
                        utils.base64_url_encode(rs.enterpriseKeys.rsaEncryptedPrivateKey)
                if rs.enterpriseKeys.eccEncryptedPrivateKey:
                    keys['ecc_public_key'] = utils.base64_url_encode(rs.enterpriseKeys.eccPublicKey)
                    keys['ecc_encrypted_private_key'] = \
                        utils.base64_url_encode(rs.enterpriseKeys.eccEncryptedPrivateKey)
            if 'rsa_encrypted_private_key' not in keys:
                rsa_private, rsa_public = crypto.generate_rsa_key()
                rsa_private_key = crypto.unload_rsa_private_key(rsa_private)
                rsa_encrypted_private_key = crypto.encrypt_aes_v2(rsa_private_key, self._enterprise.tree_key)
                rsa_public_key = crypto.unload_rsa_public_key(rsa_public)
                rq = proto.EnterpriseKeyPairRequest()
                rq.enterprisePublicKey = rsa_public_key
                rq.encryptedEnterprisePrivateKey = rsa_encrypted_private_key
                rq.keyType = proto.RSA
                api.communicate_rest(params, rq, 'enterprise/set_enterprise_key_pair')
                self._enterprise._rsa_key = rsa_private_key
                keys['rsa_public_key'] = utils.base64_url_encode(rsa_public_key)
                keys['rsa_encrypted_private_key'] = utils.base64_url_encode(rsa_encrypted_private_key)

            if 'ecc_encrypted_private_key' not in keys:
                ec_private, ec_public = crypto.generate_ec_key()
                ec_private_key = crypto.unload_ec_private_key(ec_private)
                ec_encrypted_private_key = crypto.encrypt_aes_v2(ec_private_key, self._enterprise.tree_key)
                ec_public_key = crypto.unload_ec_public_key(ec_public)
                rq = proto.EnterpriseKeyPairRequest()
                rq.enterprisePublicKey = ec_public_key
                rq.encryptedEnterprisePrivateKey = ec_encrypted_private_key
                rq.keyType = proto.ECC
                api.communicate_rest(params, rq, 'enterprise/set_enterprise_key_pair')
                keys['ecc_public_key'] = utils.base64_url_encode(ec_public_key)
                keys['ecc_encrypted_private_key'] = utils.base64_url_encode(ec_encrypted_private_key)

            params.enterprise['keys'] = keys
        entities = set()
        while True:
            rq = proto.EnterpriseDataRequest()
            if self._continuationToken:
                rq.continuationToken = self._continuationToken
            rs = api.communicate_rest(params, rq, 'enterprise/get_enterprise_data_for_user',
                                      rs_type=proto.EnterpriseDataResponse)

            if rs.cacheStatus == proto.CLEAR:
                for d in self._data_types.values():
                    d.clear(params)

            if not self._enterprise.enterprise_name and rs.generalData:
                self._enterprise._enterprise_name = rs.generalData.enterpriseName
                params.enterprise['enterprise_name'] = self._enterprise.enterprise_name
                if rs.generalData.distributor:
                    params.enterprise['distributor'] = True

            for ed in rs.data:
                entities.add(ed.entity)
                parser = self._data_types.get(ed.entity)
                if parser:
                    parser.parse(params, ed)

            self._continuationToken = rs.continuationToken
            if not rs.hasMore:
                break
        if proto.MANAGED_NODES in entities:
            self.load_missing_role_keys(params)
        if not entities.isdisjoint([proto.MANAGED_NODES, proto.NODES, proto.ROLE_USERS]):
            if 'user_root_nodes' in params.enterprise:
                del params.enterprise['user_root_nodes']
            if 'user_managed_nodes' in params.enterprise:
                del params.enterprise['user_managed_nodes']

    @staticmethod
    def load_missing_role_keys(params):   # type: (KeeperParams) -> None
        nodes = set()
        if 'managed_nodes' in params.enterprise:
            for mn in params.enterprise['managed_nodes']:
                nodes.add(mn['role_id'])
        if len(nodes) > 0:
            roles = set()
            if 'role_keys' in params.enterprise:
                for rk in params.enterprise['role_keys']:
                    roles.add(rk['role_id'])
            if 'role_keys2' in params.enterprise:
                for rk in params.enterprise['role_keys2']:
                    roles.add(rk['role_id'])
            nodes.difference_update(roles)
        if len(nodes) > 0:
            rq = proto.GetEnterpriseDataKeysRequest()
            rq.roleId.extend(nodes)
            rs = api.communicate_rest(params, rq, 'enterprise/get_enterprise_data_keys',
                                      rs_type=proto.GetEnterpriseDataKeysResponse)
            if len(rs.roleKey) > 0:
                if 'role_keys' not in params.enterprise:
                    params.enterprise['role_keys'] = []
                for rk1 in rs.roleKey:
                    params.enterprise['role_keys'].append({
                        'role_id': rk1.roleId,
                        'encrypted_key': utils.base64_url_encode(rk1.encryptedKey),
                        'key_type': _to_key_type(rk1.keyType)
                    })

            if len(rs.reEncryptedRoleKey) > 0:
                if 'role_keys2' not in params.enterprise:
                    params.enterprise['role_keys2'] = []
                for rk2 in rs.reEncryptedRoleKey:
                    params.enterprise['role_keys2'].append({
                        'role_id': rk2.role_id,
                        'role_key': utils.base64_url_encode(rk2.encryptedRoleKey),
                    })


class _EnterpriseDataParser(abc.ABC):
    def __init__(self, enterprise):    # type: (EnterpriseInfo) -> None
        self.enterprise = enterprise

    @abc.abstractmethod
    def parse(self, params, enterprise_data, **kwargs):  # type: (KeeperParams, proto.EnterpriseData, dict) -> None
        pass

    @abc.abstractmethod
    def get_entity_type(self):
        pass

    @abc.abstractmethod
    def get_keeper_entity_name(self):  # type: () -> str
        pass

    @abc.abstractmethod
    def to_keeper_entity(self, proto_entity, keeper_entity):
        pass

    def get_entities(self, params, create_if_absent=True):  # type: (KeeperParams, bool) -> Optional[List]
        name = self.get_keeper_entity_name()
        if name not in params.enterprise:
            if not create_if_absent:
                return None
            params.enterprise[name] = []
        return params.enterprise[name]

    def clear(self, params):  # type: (KeeperParams) -> None
        entities = self.get_entities(params, create_if_absent=False)
        if entities:
            entities.clear()


class _EnterpriseEntity(_EnterpriseDataParser):
    def __init__(self, enterprise):  # type: (EnterpriseInfo) -> None
        super(_EnterpriseEntity, self).__init__(enterprise)
        self._links = []     # type: List[Tuple[str, _EnterpriseLink]]

    @abc.abstractmethod
    def get_keeper_entity_id(self, proto_entity):  # type: (dict) -> any
        pass

    @abc.abstractmethod
    def get_proto_entity_id(self, proto_entity):  # type: (message.Message) -> any
        pass

    @staticmethod
    def fix_data(d):
        idx = d.rfind(b'}')
        if idx < len(d) - 1:
            d = d[:idx+1]
        return d

    def register_link(self, keeper_entity_id_name, parser):  # type: (str, _EnterpriseDataParser) -> None
        if isinstance(parser, _EnterpriseLink):
            self._links.append((keeper_entity_id_name, parser))

    def parse(self, params, enterprise_data, **kwargs):  # type: (KeeperParams, proto.EnterpriseData, dict) -> None
        if not enterprise_data.data:
            return

        entities = self.get_entities(params)
        entity_map = {self.get_keeper_entity_id(x): x for x in entities}
        entity_type = self.get_entity_type()
        deleted_entities = set()
        for entityData in enterprise_data.data:
            entity = entity_type()
            entity.ParseFromString(entityData)
            entity_id = self.get_proto_entity_id(entity)
            if enterprise_data.delete:
                if entity_id in entity_map:
                    entity_map.pop(entity_id)
                    deleted_entities.add(entity_id)
            else:
                keeper_entity = entity_map.get(entity_id)
                if not keeper_entity:
                    keeper_entity = {}
                    entity_map[entity_id] = keeper_entity
                self.to_keeper_entity(entity, keeper_entity)

        entities.clear()
        entities.extend(entity_map.values())
        if len(deleted_entities) > 0:
            for keeper_entity_id_name, link in self._links:
                link.cascade_delete(params, keeper_entity_id_name, deleted_entities)


class _EnterpriseLink(_EnterpriseDataParser):
    @abc.abstractmethod
    def get_keeper_entity1_id(self, proto_entity):  # type: (dict) -> any
        pass

    @abc.abstractmethod
    def get_keeper_entity2_id(self, proto_entity):  # type: (dict) -> any
        pass

    @abc.abstractmethod
    def get_proto_entity1_id(self, proto_entity):  # type: (message.Message) -> any
        pass

    @abc.abstractmethod
    def get_proto_entity2_id(self, proto_entity):  # type: (message.Message) -> any
        pass

    def cascade_delete(self, params, keeper_entity_id, deleted_entities):   # type: (KeeperParams, str, Set) -> None
        entities = self.get_entities(params, create_if_absent=False)
        if not entities:
            return
        to_keep = [x for x in entities if keeper_entity_id not in x or x[keeper_entity_id] not in deleted_entities]
        if len(to_keep) < len(entities):
            entities.clear()
            entities.extend(to_keep)

    def parse(self, params, enterprise_data, **kwargs):  # type: (KeeperParams, proto.EnterpriseData, dict) -> None
        entities = self.get_entities(params)
        entity_map = {
            '{0}:{1}'.format(self.get_keeper_entity1_id(x), self.get_keeper_entity2_id(x)): x for x in entities
        }
        entity_type = self.get_entity_type()
        for entityData in enterprise_data.data:
            entity = entity_type()
            entity.ParseFromString(entityData)
            entity1_id = self.get_proto_entity1_id(entity)
            entity2_id = self.get_proto_entity2_id(entity)
            key = '{0}:{1}'.format(entity1_id, entity2_id)
            if enterprise_data.delete:
                if key in entity_map:
                    entity_map.pop(key)
            else:
                keeper_entity = entity_map.get(key)
                if not keeper_entity:
                    keeper_entity = {}
                    entity_map[key] = keeper_entity
                self.to_keeper_entity(entity, keeper_entity)

        entities.clear()
        entities.extend(entity_map.values())

    def get_entities(self, params, create_if_absent=True):  # type: (KeeperParams, bool) -> Optional[List]
        name = self.get_keeper_entity_name()
        if name not in params.enterprise:
            if not create_if_absent:
                return None
            params.enterprise[name] = []
        return params.enterprise[name]


def _set_or_remove(obj, key, value):  # type: (dict, str, any) -> None
    if value is not None:
        obj[key] = value
    else:
        if key in obj:
            obj.pop(key)


class _EnterpriseNodeEntity(_EnterpriseEntity):
    def to_keeper_entity(self, proto_entity, keeper_entity):  # type: (proto.Node, dict) -> None
        _set_or_remove(keeper_entity, 'node_id', proto_entity.nodeId)
        _set_or_remove(keeper_entity, 'parent_id', proto_entity.parentId if proto_entity.parentId > 0 else None)
        _set_or_remove(keeper_entity, 'bridge_id', proto_entity.bridgeId if proto_entity.bridgeId > 0 else None)
        _set_or_remove(keeper_entity, 'scim_id', proto_entity.scimId if proto_entity.scimId > 0 else None)
        _set_or_remove(keeper_entity, 'license_id', proto_entity.licenseId if proto_entity.licenseId > 0 else None)
        _set_or_remove(keeper_entity, 'encrypted_data', proto_entity.encryptedData)
        _set_or_remove(keeper_entity, 'duo_enabled', True if proto_entity.duoEnabled else None)
        _set_or_remove(keeper_entity, 'rsa_enabled', True if proto_entity.rsaEnabled else None)
        _set_or_remove(keeper_entity, 'sso_service_provider_id',
                       proto_entity.ssoServiceProviderId if proto_entity.ssoServiceProviderId > 0 else None)
        _set_or_remove(keeper_entity, 'restrict_visibility',
                       proto_entity.restrictVisibility if proto_entity.restrictVisibility else None)

        data = {}
        if 'encrypted_data' in keeper_entity:
            try:
                data_json = api.decrypt_data(keeper_entity['encrypted_data'], self.enterprise.tree_key)
                data_json = self.fix_data(data_json)
                data.update(json.loads(data_json.decode('utf-8')))
            except Exception as e:
                logging.warning('Decrypt encryption data error: %s', e)
        elif 'parent_id' not in keeper_entity:
            data['displayname'] = self.enterprise.enterprise_name
        keeper_entity['data'] = data

    def get_keeper_entity_id(self, entity):  # type: (dict) -> any
        return entity.get('node_id')

    def get_proto_entity_id(self, entity):  # type: (proto.Node) -> any
        return entity.nodeId

    def get_entity_type(self):
        return proto.Node

    def get_keeper_entity_name(self):  # type: () -> str
        return 'nodes'


class _EnterpriseUserEntity(_EnterpriseEntity):
    def to_keeper_entity(self, proto_entity, keeper_entity):  # type: (proto.User, dict) -> None
        _set_or_remove(keeper_entity, 'enterprise_user_id', self.get_proto_entity_id(proto_entity))
        _set_or_remove(keeper_entity, 'node_id', proto_entity.nodeId)
        _set_or_remove(keeper_entity, 'username', proto_entity.username)
        _set_or_remove(keeper_entity, 'encrypted_data', proto_entity.encryptedData)
        _set_or_remove(keeper_entity, 'key_type', proto_entity.keyType)
        _set_or_remove(keeper_entity, 'status', proto_entity.status)
        _set_or_remove(keeper_entity, 'lock', proto_entity.lock)
        _set_or_remove(keeper_entity, 'user_id', proto_entity.userId)
        _set_or_remove(keeper_entity, 'account_share_expiration',
                       proto_entity.accountShareExpiration if proto_entity.accountShareExpiration > 0 else None)
        _set_or_remove(keeper_entity, 'full_name', proto_entity.fullName if proto_entity.fullName else None)
        _set_or_remove(keeper_entity, 'job_title', proto_entity.jobTitle if proto_entity.jobTitle else None)
        data = {}
        encrypted_data = keeper_entity.get('encrypted_data')
        if encrypted_data:
            if keeper_entity.get('key_type') == 'no_key':
                data['displayname'] = encrypted_data
            else:
                try:
                    data_json = api.decrypt_data(encrypted_data, self.enterprise.tree_key)
                    data_json = self.fix_data(data_json)
                    data.update(json.loads(data_json.decode('utf-8')))
                except Exception as e:
                    logging.warning('Decrypt User data error: %s', e)
        elif 'full_name' in keeper_entity:
            data['displayname'] = keeper_entity['full_name']
        keeper_entity['data'] = data

    def get_keeper_entity_id(self, entity):  # type: (dict) -> any
        return entity.get('enterprise_user_id')

    def get_proto_entity_id(self, entity):  # type: (proto.User) -> any
        return entity.enterpriseUserId

    def get_entity_type(self):
        return proto.User

    def get_keeper_entity_name(self):  # type: () -> str
        return 'users'


class _EnterpriseTeamEntity(_EnterpriseEntity):
    def to_keeper_entity(self, proto_entity, keeper_entity):  # type: (proto.Team, dict) -> None
        _set_or_remove(keeper_entity, 'team_uid', self.get_proto_entity_id(proto_entity))
        _set_or_remove(keeper_entity, 'name', proto_entity.name)
        _set_or_remove(keeper_entity, 'node_id', proto_entity.nodeId)
        _set_or_remove(keeper_entity, 'restrict_edit', proto_entity.restrictEdit)
        _set_or_remove(keeper_entity, 'restrict_sharing', proto_entity.restrictShare)
        _set_or_remove(keeper_entity, 'restrict_view', proto_entity.restrictView)
        _set_or_remove(keeper_entity, 'encrypted_data', proto_entity.encryptedData)
        _set_or_remove(keeper_entity, 'encrypted_team_key', proto_entity.encryptedTeamKey)

    def get_keeper_entity_id(self, entity):  # type: (dict) -> any
        return entity.get('team_uid')

    def get_proto_entity_id(self, entity):  # type: (proto.Team) -> any
        return utils.base64_url_encode(entity.teamUid)

    def get_entity_type(self):
        return proto.Team

    def get_keeper_entity_name(self):  # type: () -> str
        return 'teams'


class _EnterpriseRoleEntity(_EnterpriseEntity):
    def to_keeper_entity(self, proto_entity, keeper_entity):  # type: (proto.Role, dict) -> None
        _set_or_remove(keeper_entity, 'role_id', self.get_proto_entity_id(proto_entity))
        _set_or_remove(keeper_entity, 'node_id', proto_entity.nodeId)
        _set_or_remove(keeper_entity, 'encrypted_data', proto_entity.encryptedData)
        _set_or_remove(keeper_entity, 'visible_below', proto_entity.visibleBelow)
        _set_or_remove(keeper_entity, 'new_user_inherit', proto_entity.newUserInherit)
        _set_or_remove(keeper_entity, 'key_type', proto_entity.keyType)
        _set_or_remove(keeper_entity, 'role_type', proto_entity.roleType)
        data = {}
        encrypted_data = keeper_entity.get('encrypted_data')
        if encrypted_data:
            try:
                data_json = api.decrypt_data(encrypted_data, self.enterprise.tree_key)
                data_json = self.fix_data(data_json)
                data.update(json.loads(data_json.decode('utf-8')))
                if proto_entity.roleType == "pool_manager":
                    data['displayname'] = 'MSP Subscription Manager'
            except Exception as e:
                logging.warning('Decrypt encryption data error: %s', e)
        keeper_entity['data'] = data

    def get_keeper_entity_id(self, entity):  # type: (dict) -> any
        return entity.get('role_id')

    def get_proto_entity_id(self, entity):  # type: (proto.Role) -> any
        return entity.roleId

    def get_entity_type(self):
        return proto.Role

    def get_keeper_entity_name(self):  # type: () -> str
        return 'roles'


class _EnterpriseLicenseEntity(_EnterpriseEntity):
    def to_keeper_entity(self, proto_entity, keeper_entity):  # type: (proto.License, dict) -> None
        _set_or_remove(keeper_entity, 'enterprise_license_id', proto_entity.enterpriseLicenseId)
        _set_or_remove(keeper_entity, 'name', proto_entity.name)
        _set_or_remove(keeper_entity, 'paid', proto_entity.paid)
        _set_or_remove(keeper_entity, 'number_of_seats', proto_entity.numberOfSeats)
        _set_or_remove(keeper_entity, 'expiration', proto_entity.expiration)
        _set_or_remove(keeper_entity, 'license_key_id',
                       proto_entity.licenseKeyId if proto_entity.licenseKeyId > 0 else None)
        _set_or_remove(keeper_entity, 'product_type_id',
                       proto_entity.productTypeId if proto_entity.productTypeId > 0 else None)
        _set_or_remove(keeper_entity, 'seats_allocated', proto_entity.seatsAllocated)
        _set_or_remove(keeper_entity, 'seats_pending', proto_entity.seatsPending)
        _set_or_remove(keeper_entity, 'tier', proto_entity.tier)
        _set_or_remove(keeper_entity, 'file_plan',
                       proto_entity.filePlanTypeId if proto_entity.filePlanTypeId > 0 else None)
        _set_or_remove(keeper_entity, 'max_gb',
                       int(proto_entity.maxBytes / 1024 / 1024 / 1024) if proto_entity.filePlanTypeId > 0 else None)
        _set_or_remove(keeper_entity, 'storage_expiration',
                       proto_entity.storageExpiration if proto_entity.storageExpiration > 0 else None)
        _set_or_remove(keeper_entity, 'lic_status', proto_entity.licenseStatus)
        _set_or_remove(keeper_entity, 'distributor', proto_entity.distributor)

        if proto_entity.mspPool:
            msp_pool = [{
                'product_id': x.productId,
                'seats': x.seats,
                'availableSeats': x.availableSeats,
                'stash': x.stash
            } for x in proto_entity.mspPool]
            _set_or_remove(keeper_entity, 'msp_pool', msp_pool)

        if proto_entity.managedBy and proto_entity.managedBy.enterpriseId > 0:
            _set_or_remove(keeper_entity, 'managed_by', {
                'enterprise_id': proto_entity.managedBy.enterpriseId,
                'enterprise_name': proto_entity.managedBy.enterpriseName,
            })

        if proto_entity.addOns:
            _set_or_remove(keeper_entity, 'add_ons', [{
                'name': x.name,
                'enabled': x.enabled,
                'is_trial': x.isTrial,
                'created': x.created,
                'expiration': x.expiration,
            } for x in proto_entity.addOns])

        if proto_entity.mspPermits.restricted:
            _set_or_remove(keeper_entity, 'msp_permits', {
                'allow_unlimited_licenses': proto_entity.mspPermits.allowUnlimitedLicenses,
                'allowed_mc_products': [x for x in proto_entity.mspPermits.allowedMcProducts],
                'allowed_add_ons': [x for x in proto_entity.mspPermits.allowedAddOns],
                'max_file_plan_type': proto_entity.mspPermits.maxFilePlanType,
                'mc_defaults': [{
                    'mc_product': x.mcProduct,
                    'add_ons': [a for a in x.addOns],
                    'file_plan_type': x.filePlanType,
                } for x in proto_entity.mspPermits.mcDefaults]
            })

        _set_or_remove(keeper_entity, 'next_billing_date',
                       proto_entity.nextBillingDate if proto_entity.nextBillingDate > 0 else None)

    def get_keeper_entity_id(self, entity):  # type: (dict) -> any
        return entity.get('enterprise_license_id')

    def get_proto_entity_id(self, entity):  # type: (proto.License) -> any
        return entity.enterpriseLicenseId

    def get_entity_type(self):
        return proto.License

    def get_keeper_entity_name(self):  # type: () -> str
        return 'licenses'


class _EnterpriseQueuedTeamEntity(_EnterpriseEntity):
    def to_keeper_entity(self, proto_entity, keeper_entity):  # type: (proto.QueuedTeam, dict) -> None
        _set_or_remove(keeper_entity, 'team_uid', self.get_proto_entity_id(proto_entity))
        _set_or_remove(keeper_entity, 'name', proto_entity.name)
        _set_or_remove(keeper_entity, 'node_id', proto_entity.nodeId)
        _set_or_remove(keeper_entity, 'encrypted_data', proto_entity.encryptedData)

    def get_keeper_entity_id(self, entity):  # type: (dict) -> any
        return entity.get('team_uid')

    def get_proto_entity_id(self, entity):  # type: (proto.QueuedTeam) -> any
        return utils.base64_url_encode(entity.teamUid)

    def get_entity_type(self):
        return proto.QueuedTeam

    def get_keeper_entity_name(self):  # type: () -> str
        return 'queued_teams'


class _EnterpriseScimEntity(_EnterpriseEntity):
    def to_keeper_entity(self, proto_entity, keeper_entity):  # type: (proto.Scim, dict) -> None
        _set_or_remove(keeper_entity, 'scim_id', self.get_proto_entity_id(proto_entity))
        _set_or_remove(keeper_entity, 'node_id', proto_entity.nodeId)
        _set_or_remove(keeper_entity, 'status', proto_entity.status)
        _set_or_remove(keeper_entity, 'last_synced', proto_entity.lastSynced if proto_entity.lastSynced > 0 else None)
        _set_or_remove(keeper_entity, 'role_prefix', proto_entity.rolePrefix)
        _set_or_remove(keeper_entity, 'unique_groups', proto_entity.uniqueGroups)

    def get_keeper_entity_id(self, entity):  # type: (dict) -> any
        return entity.get('scim_id')

    def get_proto_entity_id(self, entity):  # type: (proto.Scim) -> any
        return entity.scimId

    def get_entity_type(self):
        return proto.Scim

    def get_keeper_entity_name(self):  # type: () -> str
        return 'scims'


class _EnterpriseTeamUserEntity(_EnterpriseLink):
    def to_keeper_entity(self, proto_entity, keeper_entity):  # type: (proto.TeamUser, dict) -> None
        _set_or_remove(keeper_entity, 'team_uid', self.get_proto_entity1_id(proto_entity))
        _set_or_remove(keeper_entity, 'enterprise_user_id', proto_entity.enterpriseUserId)
        user_type = 0 if proto_entity.userType == 'USER' else 1 if proto_entity.userType == 'ADMIN' else 2
        _set_or_remove(keeper_entity, 'user_type', user_type)

    def get_keeper_entity1_id(self, entity):  # type: (dict) -> any
        return entity.get('team_uid')

    def get_keeper_entity2_id(self, entity):  # type: (dict) -> any
        return entity.get('enterprise_user_id')

    def get_proto_entity1_id(self, entity):  # type: (proto.TeamUser) -> any
        return utils.base64_url_encode(entity.teamUid)

    def get_proto_entity2_id(self, entity):  # type: (proto.TeamUser) -> any
        return entity.enterpriseUserId

    def get_entity_type(self):
        return proto.TeamUser

    def get_keeper_entity_name(self):  # type: () -> str
        return 'team_users'


class _EnterpriseRoleUserEntity(_EnterpriseLink):
    def to_keeper_entity(self, proto_entity, keeper_entity):  # type: (proto.RoleUser, dict) -> None
        _set_or_remove(keeper_entity, 'role_id', self.get_proto_entity1_id(proto_entity))
        _set_or_remove(keeper_entity, 'enterprise_user_id', proto_entity.enterpriseUserId)

    def get_keeper_entity1_id(self, entity):  # type: (dict) -> any
        return entity.get('role_id')

    def get_keeper_entity2_id(self, entity):  # type: (dict) -> any
        return entity.get('enterprise_user_id')

    def get_proto_entity1_id(self, entity):  # type: (proto.RoleUser) -> any
        return entity.roleId

    def get_proto_entity2_id(self, entity):  # type: (proto.RoleUser) -> any
        return entity.enterpriseUserId

    def get_entity_type(self):
        return proto.RoleUser

    def get_keeper_entity_name(self):  # type: () -> str
        return 'role_users'


class _EnterpriseRoleTeamEntity(_EnterpriseLink):
    def to_keeper_entity(self, proto_entity, keeper_entity):  # type: (proto.RoleTeam, dict) -> None
        _set_or_remove(keeper_entity, 'role_id', self.get_proto_entity1_id(proto_entity))
        _set_or_remove(keeper_entity, 'team_uid', self.get_proto_entity2_id(proto_entity))

    def get_keeper_entity1_id(self, entity):  # type: (dict) -> any
        return entity.get('role_id')

    def get_keeper_entity2_id(self, entity):  # type: (dict) -> any
        return entity.get('team_uid')

    def get_proto_entity1_id(self, entity):  # type: (proto.RoleTeam) -> any
        return entity.role_id

    def get_proto_entity2_id(self, entity):  # type: (proto.RoleTeam) -> any
        return utils.base64_url_encode(entity.teamUid)

    def get_entity_type(self):
        return proto.RoleTeam

    def get_keeper_entity_name(self):  # type: () -> str
        return 'role_teams'


class _EnterpriseManagedNodeEntity(_EnterpriseLink):
    def to_keeper_entity(self, proto_entity, keeper_entity):  # type: (proto.ManagedNode, dict) -> None
        _set_or_remove(keeper_entity, 'role_id', self.get_proto_entity1_id(proto_entity))
        _set_or_remove(keeper_entity, 'managed_node_id', self.get_proto_entity2_id(proto_entity))
        _set_or_remove(keeper_entity, 'cascade_node_management', proto_entity.cascadeNodeManagement)

    def get_keeper_entity1_id(self, entity):  # type: (dict) -> any
        return entity.get('role_id')

    def get_keeper_entity2_id(self, entity):  # type: (dict) -> any
        return entity.get('managed_node_id')

    def get_proto_entity1_id(self, entity):  # type: (proto.ManagedNode) -> any
        return entity.roleId

    def get_proto_entity2_id(self, entity):  # type: (proto.ManagedNode) -> any
        return entity.managedNodeId

    def get_entity_type(self):
        return proto.ManagedNode

    def get_keeper_entity_name(self):  # type: () -> str
        return 'managed_nodes'


class _EnterpriseRolePrivilegeEntity(_EnterpriseEntity):
    def to_keeper_entity(self, proto_entity, keeper_entity):  # type: (proto.RolePrivilege, dict) -> None
        _set_or_remove(keeper_entity, 'role_id', proto_entity.roleId)
        _set_or_remove(keeper_entity, 'managed_node_id', proto_entity.managedNodeId)
        _set_or_remove(keeper_entity, 'privilege', proto_entity.privilegeType)

    def get_keeper_entity_id(self, entity):  # type: (dict) -> any
        return '{0}:{1}:{2}'.format(entity.get('role_id'), entity.get('managed_node_id'), entity.get('privilege'))

    def get_proto_entity_id(self, entity):  # type: (proto.RolePrivilege) -> any
        return '{0}:{1}:{2}'.format(entity.roleId, entity.managedNodeId, entity.privilegeType)

    def get_entity_type(self):
        return proto.RolePrivilege

    def get_keeper_entity_name(self):  # type: () -> str
        return 'role_privileges'


class _EnterpriseRoleEnforcements(_EnterpriseDataParser):
    def parse(self, params, enterprise_data, **kwargs):  # type: (KeeperParams, proto.EnterpriseData, dict) -> None
        entities = self.get_entities(params)
        entity_map = {x['role_id']: x for x in entities}
        entity_type = self.get_entity_type()
        for entityData in enterprise_data.data:
            entity = entity_type()
            entity.ParseFromString(entityData)
            role_id = entity.roleId
            enforcement_type = entity.enforcementType
            if enterprise_data.delete:
                if role_id in entity_map:
                    enforcements = entity_map[role_id]['enforcements']
                    if enforcement_type in enforcements:
                        enforcements.pop(enforcement_type)
            else:
                keeper_entity = entity_map.get(role_id)
                if not keeper_entity:
                    keeper_entity = {
                        'role_id': role_id,
                        'enforcements': {}
                    }
                    entity_map[role_id] = keeper_entity
                enforcements = keeper_entity['enforcements']
                enforcements[enforcement_type] = entity.value

        entities.clear()
        entities.extend(entity_map.values())

    def get_entity_type(self):
        return proto.RoleEnforcement

    def get_keeper_entity_name(self):  # type: () -> str
        return 'role_enforcements'

    def to_keeper_entity(self, proto_entity, keeper_entity):
        pass


class _EnterpriseManagedCompanyEntity(_EnterpriseEntity):
    def to_keeper_entity(self, proto_entity, keeper_entity):  # type: (proto.ManagedCompany, dict) -> None
        _set_or_remove(keeper_entity, 'mc_enterprise_id', proto_entity.mcEnterpriseId)
        _set_or_remove(keeper_entity, 'mc_enterprise_name', proto_entity.mcEnterpriseName)
        _set_or_remove(keeper_entity, 'msp_node_id', proto_entity.mspNodeId)
        _set_or_remove(keeper_entity, 'number_of_seats', proto_entity.numberOfSeats)
        _set_or_remove(keeper_entity, 'number_of_users', proto_entity.numberOfUsers)
        _set_or_remove(keeper_entity, 'product_id', proto_entity.productId)
        _set_or_remove(keeper_entity, 'paused', proto_entity.isExpired)
        _set_or_remove(keeper_entity, 'tree_key', proto_entity.treeKey if proto_entity.treeKey else None)
        _set_or_remove(keeper_entity, 'tree_key_role', proto_entity.tree_key_role)
        _set_or_remove(keeper_entity, 'file_plan_type', proto_entity.filePlanType)
        _set_or_remove(keeper_entity, 'add_ons', [{
            'name': x.name,
            'seats': x.seats,
            'enabled': x.enabled,
            'is_trial': x.isTrial,
            'created': x.created,
            'expiration': x.expiration,
            'activation_time': x.activationTime,
            'included_in_product': x.includedInProduct,
        } for x in proto_entity.addOns])

    def get_keeper_entity_id(self, entity):  # type: (dict) -> any
        return entity.get('mc_enterprise_id')

    def get_proto_entity_id(self, entity):  # type: (proto.ManagedCompany) -> any
        return entity.mcEnterpriseId

    def get_entity_type(self):
        return proto.ManagedCompany

    def get_keeper_entity_name(self):  # type: () -> str
        return 'managed_companies'


class _EnterpriseQueuedTeamUserEntity(_EnterpriseDataParser):
    def parse(self, params, enterprise_data, **kwargs):  # type: (KeeperParams, proto.EnterpriseData, dict) -> None
        entities = self.get_entities(params)
        entity_map = {x['team_uid']: x for x in entities}
        entity_type = self.get_entity_type()
        for entityData in enterprise_data.data:
            entity = entity_type()
            entity.ParseFromString(entityData)
            team_uid = utils.base64_url_encode(entity.teamUid)
            if enterprise_data.delete:
                if team_uid in entity_map:
                    users = entity_map[team_uid]['users']     # type: set
                    users.difference_update(entity.users)
            else:
                keeper_entity = entity_map.get(team_uid)
                if not keeper_entity:
                    keeper_entity = {
                        'team_uid': team_uid,
                        'users': set()
                    }
                    entity_map[team_uid] = keeper_entity
                users = keeper_entity['users']
                users.update(entity.users)

        entities.clear()
        entities.extend(entity_map.values())

    def get_entity_type(self):
        return proto.QueuedTeamUser

    def get_keeper_entity_name(self):  # type: () -> str
        return 'queued_team_users'

    def to_keeper_entity(self, proto_entity, keeper_entity):
        pass


class _EnterpriseAdminApprovalRequestEntity(_EnterpriseEntity):
    def to_keeper_entity(self, proto_entity, keeper_entity):
        # type: (proto.DeviceRequestForAdminApproval, dict) -> None
        _set_or_remove(keeper_entity, 'enterprise_user_id', proto_entity.enterpriseUserId)
        _set_or_remove(keeper_entity, 'encrypted_device_token',
                       utils.base64_url_encode(proto_entity.encryptedDeviceToken))
        _set_or_remove(keeper_entity, 'device_id', proto_entity.deviceId)
        _set_or_remove(keeper_entity, 'device_public_key', utils.base64_url_encode(proto_entity.devicePublicKey))
        _set_or_remove(keeper_entity, 'device_name', proto_entity.deviceName)
        _set_or_remove(keeper_entity, 'client_version', proto_entity.clientVersion)
        _set_or_remove(keeper_entity, 'device_type', proto_entity.deviceType)
        _set_or_remove(keeper_entity, 'date', proto_entity.date)
        _set_or_remove(keeper_entity, 'ip_address', proto_entity.ipAddress)
        _set_or_remove(keeper_entity, 'location', proto_entity.location)
        _set_or_remove(keeper_entity, 'email', proto_entity.email)

    def get_keeper_entity_id(self, entity):  # type: (dict) -> any
        return '{0}:{1}'.format(entity.get('enterprise_user_id'), entity.get('device_id'))

    def get_proto_entity_id(self, entity):  # type: (proto.DeviceRequestForAdminApproval) -> any
        return '{0}:{1}'.format(entity.enterpriseUserId, entity.deviceId)

    def get_entity_type(self):
        return proto.DeviceRequestForAdminApproval

    def get_keeper_entity_name(self):  # type: () -> str
        return 'devices_request_for_admin_approval'


class _EnterpriseSsoServiceEntity(_EnterpriseEntity):
    def to_keeper_entity(self, proto_entity, keeper_entity):  # type: (proto.SsoService, dict) -> None
        _set_or_remove(keeper_entity, 'sso_service_provider_id', self.get_proto_entity_id(proto_entity))
        _set_or_remove(keeper_entity, 'node_id', proto_entity.nodeId)
        _set_or_remove(keeper_entity, 'name', proto_entity.name)
        _set_or_remove(keeper_entity, 'sp_url', proto_entity.sp_url)
        _set_or_remove(keeper_entity, 'invite_new_users', proto_entity.inviteNewUsers)
        _set_or_remove(keeper_entity, 'active', proto_entity.active)
        _set_or_remove(keeper_entity, 'is_cloud', proto_entity.isCloud)

    def get_keeper_entity_id(self, entity):  # type: (dict) -> any
        return entity.get('sso_service_provider_id')

    def get_proto_entity_id(self, entity):  # type: (proto.SsoService) -> any
        return entity.ssoServiceProviderId

    def get_entity_type(self):
        return proto.SsoService

    def get_keeper_entity_name(self):  # type: () -> str
        return 'sso_services'


class _EnterpriseBridgeEntity(_EnterpriseEntity):
    def to_keeper_entity(self, proto_entity, keeper_entity):  # type: (proto.Bridge, dict) -> None
        _set_or_remove(keeper_entity, 'bridge_id', self.get_proto_entity_id(proto_entity))
        _set_or_remove(keeper_entity, 'node_id', proto_entity.nodeId)
        _set_or_remove(keeper_entity, 'wan_ip_enforcement', proto_entity.wanIpEnforcement)
        _set_or_remove(keeper_entity, 'lan_ip_enforcement', proto_entity.lanIpEnforcement)
        _set_or_remove(keeper_entity, 'status', proto_entity.status)

    def get_keeper_entity_id(self, entity):  # type: (dict) -> any
        return entity.get('bridge_id')

    def get_proto_entity_id(self, entity):  # type: (proto.Bridge) -> any
        return entity.bridgeId

    def get_entity_type(self):
        return proto.Bridge

    def get_keeper_entity_name(self):  # type: () -> str
        return 'bridges'


class _EnterpriseEmailProvisionEntity(_EnterpriseEntity):
    def to_keeper_entity(self, proto_entity, keeper_entity):  # type: (proto.EmailProvision, dict) -> None
        _set_or_remove(keeper_entity, 'id', self.get_proto_entity_id(proto_entity))
        _set_or_remove(keeper_entity, 'node_id', proto_entity.nodeId)
        _set_or_remove(keeper_entity, 'domain', proto_entity.domain)
        _set_or_remove(keeper_entity, 'method', proto_entity.method)

    def get_keeper_entity_id(self, entity):  # type: (dict) -> any
        return entity.get('id')

    def get_proto_entity_id(self, entity):  # type: (proto.EmailProvision) -> any
        return entity.id

    def get_entity_type(self):
        return proto.EmailProvision

    def get_keeper_entity_name(self):  # type: () -> str
        return 'email_provision'



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper SDK
# Copyright 2022 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#
from typing import Optional, List, Callable, Any

from .vault import TypedRecord, TypedField


class TypedRecordFacade:
    def __init__(self):   # type: () -> None
        self._record = None     # type: Optional[TypedRecord]

    def _get_record(self):
        return self._record

    def _set_record(self, record):
        if record is None or isinstance(record, TypedRecord):
            self._record = record
            self.load_typed_fields()
        else:
            raise ValueError('expected TypedRecord')

    def _get_title(self):
        if isinstance(self._record, TypedRecord):
            return self._record.title
        raise ValueError('typed record is not assigned')

    def _set_title(self, value):
        if isinstance(self._record, TypedRecord):
            self._record.title = value
        else:
            raise ValueError('typed record is not assigned')

    def _get_notes(self):
        if isinstance(self._record, TypedRecord):
            return self._record.notes
        raise ValueError('typed record is not assigned')

    def _set_notes(self, value):
        if isinstance(self._record, TypedRecord):
            self._record.notes = value
        else:
            raise ValueError('typed record is not assigned')

    record = property(fget=_get_record, fset=_set_record)
    title = property(fget=_get_title, fset=_set_title)
    notes = property(fget=_get_notes, fset=_set_notes)

    def load_typed_fields(self):
        pass


def string_list_getter(name):   # type: (str) -> Callable[[TypedRecordFacade], List[str]]
    def getter(obj):
        field = getattr(obj, name)
        if isinstance(field, TypedField):
            return field.value
    return getter


def string_getter(name):   # type: (str) -> Callable[[TypedRecordFacade], str]
    def getter(obj):
        field = getattr(obj, name)
        if isinstance(field, TypedField):
            return field.value[0] if len(field.value) > 0 else ''
    return getter


def string_setter(name):   # type: (str) -> Callable[[Any, Any], None]
    def setter(obj, value):
        field = getattr(obj, name)
        if isinstance(field, TypedField):
            if value:
                if len(field.value) > 0:
                    field.value[0] = value
                else:
                    field.value.append(value)
            else:
                field.value.clear()
    return setter


class FileRefRecordFacade(TypedRecordFacade):
    def __init__(self):     # type: () -> None
        super(FileRefRecordFacade, self).__init__()
        self._file_ref = None       # type: Optional[TypedField]

    def load_typed_fields(self):
        if self.record:
            self._file_ref = next((x for x in self.record.fields if x.type == 'fileRef'), None)
            if self._file_ref is None:
                self._file_ref = TypedField.new_field('fileRef', [])
                self.record.fields.append(self._file_ref)
        else:
            self._file_ref = None
        super(FileRefRecordFacade, self).load_typed_fields()

    file_ref = property(string_list_getter('_file_ref'))


class LoginRecordFacade(FileRefRecordFacade):
    def __init__(self):
        super(LoginRecordFacade, self).__init__()
        self._login = None        # type: Optional[TypedField]
        self._password = None     # type: Optional[TypedField]
        self._url = None          # type: Optional[TypedField]
        self._oneTimeCode = None  # type: Optional[TypedField]

    login = property(fget=string_getter('_login'), fset=string_setter('_login'))
    password = property(fget=string_getter('_password'), fset=string_setter('_password'))
    url = property(fget=string_getter('_url'), fset=string_setter('_url'))
    oneTimeCode = property(fget=string_getter('_oneTimeCode'), fset=string_setter('_oneTimeCode'))

    def load_typed_fields(self):
        if self.record:
            self.record.type_name = 'login'
            self._login = next((x for x in self.record.fields if x.type == 'login'), None)
            if self._login is None:
                self._login = TypedField.new_field('login', '')
                self.record.fields.append(self._login)
            self._password = next((x for x in self.record.fields if x.type == 'password'), None)
            if self._password is None:
                self._password = TypedField.new_field('password', '')
                self.record.fields.append(self._password)
            self._url = next((x for x in self.record.fields if x.type == 'url'), None)
            if self._url is None:
                self._url = TypedField.new_field('url', '')
                self.record.fields.append(self._url)
            self._oneTimeCode = next((x for x in self.record.fields if x.type == 'oneTimeCode'), None)
            if self._oneTimeCode is None:
                self._oneTimeCode = TypedField.new_field('oneTimeCode', '')
                self.record.fields.append(self._oneTimeCode)
        else:
            self._login = None
            self._password = None
            self._url = None
            self._oneTimeCode = None

        super(LoginRecordFacade, self).load_typed_fields()



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2022 Keeper Security Inc.
# Contact: commander@keepersecurity.com
#
# Example code to retrieve the password for a record
# stored in the vault.  This example also pulls configuration
# from config.json or writes the config file if it does not exist.
#
# Usage:
#    python update_password.py

import os
import json
import base64
import getpass
from typing import Optional

from keepercommander.params import KeeperParams
from keepercommander import api, vault, vault_extensions, record_management,generator
from keepercommander.commands.recordv3 import RecordEditCommand


def read_config_file(params):
    params.config_filename = os.path.join(os.path.dirname(__file__), 'config.json')
    if os.path.isfile(params.config_filename):
        with open(params.config_filename, 'r') as f:
            params.config = json.load(f)
            if 'user' in params.config:
                params.user = params.config['user']

            if 'password' in params.config:
                params.password = params.config['password']

            if 'mfa_token' in params.config:
                params.mfa_token = params.config['mfa_token']

            if 'server' in params.config:
                params.server = params.config['server']

            if 'device_id' in params.config:
                device_id = base64.urlsafe_b64decode(params.config['device_id'] + '==')
                params.rest_context.device_id = device_id


my_params = KeeperParams()
read_config_file(my_params)

while not my_params.user:
    my_params.user = getpass.getpass(prompt='User(Email): ', stream=None)

while not my_params.password:
    my_params.password = getpass.getpass(prompt='Master Password: ', stream=None)

api.sync_down(my_params)

record_uid = ''

# Find any login record

# 1. direct API
record_uid = next((x.record_uid for x in vault_extensions.find_records(my_params, record_type='login')), None)

# 2. list command
# list_command = RecordListCommand()
# login_records = list_command.execute(my_params, format='json', type='login')
# record_list = json.loads(login_records)
# if isinstance(record_list, list) and len(record_list) > 0:
#     record_uid = record_list[0]['record_uid']

# update password
password = generator.generate(20)

# 1. record-edit command
if record_uid:
    edit_command = RecordEditCommand()
    edit_command.execute(my_params, password=password, record=record_uid)
    api.sync_down(my_params)

# 2, direct API
#if record_uid:
#    record = vault.KeeperRecord.load(my_params, record_uid)
#    if isinstance(record, vault.TypedRecord):
#        password_field = record.get_typed_field('password')
#        if password_field:
#            password_field.value = [password]
#            record_management.update_record(my_params, record)
#            api.sync_down(my_params)



# -*- coding: utf-8 -*-
#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2022 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import argparse
import io
import json
import logging
import os
import re
import shutil
import sys
import tempfile
from typing import Optional, Callable, List, Iterable, Tuple

from .base import Command, RecordMixin, dump_report_data, field_to_title
from .record import find_record, RecordListCommand
from .ssh_agent import add_ssh_key, try_extract_private_key, SshAgentCommand
from ..attachment import prepare_attachment_download
from ..error import CommandError
from ..params import KeeperParams
from ..subfolder import find_folders, get_folder_path, try_resolve_path
from ..vault import TypedRecord, KeeperRecord, PasswordRecord

ssh_parser = argparse.ArgumentParser(prog='ssh',
                                     description='Establishes connection to external server using SSH. ')
ssh_parser.add_argument('--format', dest='format', action='store', choices=['table', 'csv', 'json'], default='table',
                        help='output format.')
ssh_parser.add_argument('--output', dest='output', action='store',
                        help='output file name. (ignored for table format)')
ssh_parser.add_argument('-d', '--destination', action='store', metavar='LOGIN@HOST[:PORT]',
                        help='SSH endpoint')
ssh_parser.add_argument('record', nargs='?', type=str, action='store',
                        help='record path or UID. Record types: "SSH Key", "Server"')
ssh_parser.add_argument('command', nargs='*', type=str, action='store',
                        help='Remote command')

mysql_parser = argparse.ArgumentParser(prog='mysql', description='Establishes connection to MySQL server.')
mysql_parser.add_argument('--format', dest='format', action='store', choices=['table', 'csv', 'json'], default='table',
                          help='output format.')
mysql_parser.add_argument('--output', dest='output', action='store',
                          help='output file name. (ignored for table format)')
mysql_parser.add_argument('record', nargs='?', type=str, action='store',
                          help='record path or UID. Record types: "Database"')
mysql_parser.add_argument('query', nargs='*', type=str, action='store',
                          help='SQL query')

postgres_parser = argparse.ArgumentParser(prog='postgresql',
                                          description='Establishes connection to Postgres/Redshift servers.')
postgres_parser.add_argument('--format', dest='format', action='store', choices=['table', 'csv', 'json'], default='table',
                             help='output format.')
postgres_parser.add_argument('--output', dest='output', action='store',
                             help='output file name. (ignored for table format)')
postgres_parser.add_argument('record', nargs='?', type=str, action='store',
                             help='record path or UID. Record types: "Database"')
postgres_parser.add_argument('database', nargs='?', type=str, action='store',
                             help='Postgres database name.')

rdp_parser = argparse.ArgumentParser(prog='rdp',
                                     description='Establishes RDP connection to remote Windows servers.')
rdp_parser.add_argument('record', nargs='?', type=str, action='store',
                        help='record path or UID. Record types: "Server"')


connect_parser = argparse.ArgumentParser(prog='connect', description='Establishes connection to external server')
connect_parser.add_argument('--syntax-help', dest='syntax_help', action='store_true',
                            help='display help on command format and template parameters')
connect_parser.add_argument('--format', dest='format', action='store', choices=['table', 'csv', 'json'], default='table',
                            help='output format.')
connect_parser.add_argument('--output', dest='output', action='store',
                            help='output file name. (ignored for table format)')
connect_parser.add_argument('-s', '--sort', dest='sort_by', action='store', choices=['endpoint', 'title', 'folder'],
                            help='sort output')
connect_parser.add_argument('-n', '--new', dest='new_data', action='store_true', help='request per-user data')
connect_parser.add_argument('-f', '--filter', dest='filter_by', action='store', help='filter output')
connect_parser.add_argument('endpoint', nargs='?', action='store', type=str,
                            help='endpoint name or full record path to endpoint')
connect_parser.add_argument('parameters', nargs='*', type=str, action='store',
                            help='Command parameters')

mysql = ''
postgresql = ''

endpoint_parameter_pattern = re.compile(r'\${(.+?)}')


def detect_clients():
    global mysql, postgresql
    if shutil.which('mysql'):
        mysql = 'mysql'
    if shutil.which('pgcli'):
        postgresql = 'pgcli'
    elif shutil.which('psql'):
        postgresql = 'psql'


detect_clients()


def connect_commands(commands):
    commands['ssh-agent'] = SshAgentCommand()
    commands['connect'] = ConnectCommand()
    commands['ssh'] = ConnectSshCommand()
    if mysql:
        commands['mysql'] = ConnectMysqlCommand()
    if postgresql:
        commands['postgresql'] = ConnectPostgresCommand()
    if sys.platform == 'win32':
        commands['rdp'] = ConnectRdpCommand()


def connect_command_info(aliases, command_info):
    command_info['ssh-agent'] = 'Manage SSH Agent'
    command_info[connect_parser.prog] = connect_parser.description
    command_info[ssh_parser.prog] = ssh_parser.description
    if mysql:
        command_info['mysql'] = mysql_parser.description
    if postgresql:
        command_info['postgresql'] = postgres_parser.description
        aliases['pg'] = 'postgresql'
    if sys.platform == 'win32':
        command_info['rdp'] = rdp_parser.description


class BaseConnectCommand(Command, RecordMixin):
    def __init__(self):
        super(BaseConnectCommand, self).__init__()
        self.command = ''
        self.run_at_the_end = []

    def support_extra_parameters(self):
        return True

    SHELL_SUBSTITUTION = {
        '`': r'\`',
        '$': r'\$',
        '?': r'\?',
        '*': r'\*',
        '^': r'\^',
        '(': r'\(',
        ')': r'\)'
    }

    def execute_shell(self):
        logging.debug('Executing "%s" ...', self.command)
        try:
            command = self.command.translate(str.maketrans(BaseConnectCommand.SHELL_SUBSTITUTION))
            os.system(command)
        finally:
            self.command = ''
            for cb in self.run_at_the_end:
                try:
                    cb()
                except Exception as e:
                    logging.debug(e)
            self.run_at_the_end.clear()

    def get_extra_options(self, params, record, application):
        # type: (KeeperParams, KeeperRecord, str) -> str
        record_options = BaseConnectCommand.get_custom_field(record, f'{application}:option')
        if record_options:
            temp_files = []
            record_options = BaseConnectCommand.get_command_string(params, record, record_options, temp_files)
            if temp_files:
                def remove_files():
                    for file in temp_files:
                        os.remove(file)
                self.run_at_the_end.append(remove_files)

        options = ''
        if record_options:
            options += f' {record_options}'
        if self.extra_parameters:
            options += f' {self.extra_parameters}'
        return options

    @staticmethod
    def get_parameter_value(params, record, parameter, temp_files, **kwargs):
        # type: (KeeperParams, KeeperRecord, str, list, ...) -> Optional[str]
        if parameter.startswith('file:') or parameter.startswith('body:'):
            file_name = parameter[5:]
            attachments = list(prepare_attachment_download(params, record.record_uid, file_name))
            if len(attachments) == 0:
                logging.warning('Attachment file \"%s\" not found', file_name)
                return None
            if len(attachments) > 1:
                logging.warning('More than one attachment file \"%s\" found', file_name)
                return None

            if parameter.startswith('file:'):
                prefix = (kwargs.get('endpoint') or file_name) + '.'
                with tempfile.NamedTemporaryFile(delete=False, prefix=prefix) as tf:
                    attachments[0].download_to_stream(params, tf)
                    temp_files.append(tf.name)
                    return tf.name
            else:
                with io.BytesIO() as mem:
                    attachments[0].download_to_stream(params, mem)
                    return mem.getvalue().decode('utf-8')
        else:
            return BaseConnectCommand.get_record_field(record, parameter)

    @staticmethod
    def get_command_string(params, record, template, temp_files, **kwargs):
        # type: (KeeperParams, KeeperRecord, str, list, ...) -> str or None
        command = template
        while True:
            m = endpoint_parameter_pattern.search(command)
            if not m:
                break
            p = m.group(1)
            pv = BaseConnectCommand.get_parameter_value(params, record, p, temp_files, **kwargs)
            command = command[:m.start()] + (pv or '') + command[m.end():]
        return command


class ConnectSshCommand(BaseConnectCommand):
    def get_parser(self):
        return ssh_parser

    def execute(self, params, **kwargs):
        record_name = kwargs['record'] if 'record' in kwargs else None
        if not record_name:
            ls = RecordListCommand()
            return ls.execute(params, record_type=['serverCredentials', 'sshKeys'], verbose=True, **kwargs)

        record = None     # type: Optional[KeeperRecord]
        if record_name in params.record_cache:
            record = KeeperRecord.load(params, record_name)
        else:
            rs = try_resolve_path(params, record_name)
            if rs is not None:
                folder, record_name = rs
                if folder is not None and record_name is not None:
                    folder_uid = folder.uid or ''
                    if folder_uid in params.subfolder_record_cache:
                        for uid in params.subfolder_record_cache[folder_uid]:
                            r = KeeperRecord.load(params, uid)
                            if r:
                                if r.title.lower() == record_name.lower():
                                    record = r
                                    break

        if record is None:
            ls = RecordListCommand()
            result = ls.execute(params, record_type=['serverCredentials', 'sshKeys'], format='json', verbose=True)
            if result:
                try:
                    recs = json.loads(result)
                    records = []
                    if isinstance(recs, list):
                        for rec in recs:
                            if isinstance(rec, dict):
                                if 'title' in rec:
                                    title = rec.get('title', '').strip().lower()
                                    if title == record_name.lower():
                                        records.append(rec)
                                        continue
                                if 'description' in rec:
                                    description = rec.get('description', '').lower()
                                    if description:
                                        user, sep, host = description.partition("@")
                                        if sep == '@':
                                            description = host
                                        hostname, _, _ = description.strip().partition(':')
                                        if hostname == record_name.lower():
                                            records.append(rec)
                                            continue
                    if len(records) == 1:
                        record = KeeperRecord.load(params, records[0].get('record_uid'))
                    elif len(records) > 1:
                        raise CommandError('ssh', f'More than one record found for \"{record_name}\". Please use record UID or full record path.')
                except:
                    pass

        if record is None:
            raise CommandError('ssh', 'Enter name of existing record')

        if not isinstance(record, (PasswordRecord, TypedRecord)):
            raise CommandError('ssh', f'Record \"{record.title}\" type is not supported.')

        dst = kwargs.get('destination', '')
        if dst:
            login, at, host = dst.partition('@')
            if at != '@':
                logging.warning('Destination parameter should be LOGIN@HOST[:PORT]')
                return
        else:
            login = BaseConnectCommand.get_record_field(record, 'login')
            host = BaseConnectCommand.get_record_field(record, 'host')

        if not login:
            logging.warning('Record "%s" does not have login.', record.title)
            return
        if not host:
            logging.warning('Record "%s" does not have host.', record.title)
            return

        host_name, _, port = host.partition(':')

        self.run_at_the_end.clear()

        options = self.get_extra_options(params, record, 'ssh')
        self.command = f'ssh{options} {login}@{host_name}'
        if port:
            self.command += f' -p {port}'

        pk = try_extract_private_key(params, record)
        if pk:
            private_key, passphrase = pk
            to_remove = add_ssh_key(private_key, passphrase, record.title)
            if to_remove:
                self.run_at_the_end.append(to_remove)
        else:
            password = BaseConnectCommand.get_record_field(record, 'password')
            if password:
                if shutil.which('sshpass'):
                    self.command = 'sshpass -e ' + self.command
                    os.putenv('SSHPASS', password)

                    def clear_env():
                        os.putenv('SSHPASS', '')
                    self.run_at_the_end.append(clear_env)
                else:
                    self.command += ' -o PubkeyAuthentication=no'
                    try:
                        import pyperclip
                        pyperclip.copy(password)
                        logging.info('\nPassword is copied to clipboard\n')

                        def clear_clipboard():
                            txt = pyperclip.paste()
                            if txt == password:
                                pyperclip.copy('')
                        self.run_at_the_end.append(clear_clipboard)
                    except Exception as e:
                        logging.debug(e)
                        logging.info('Failed to copy password to clipboard')

        command = kwargs.get('command')
        if isinstance(command, list) and len(command) > 0:
            self.command += ' -- ' + ' '.join(command)

        logging.info('Connecting to "%s" ...', record.title)
        self.execute_shell()


class ConnectMysqlCommand(BaseConnectCommand):
    def get_parser(self):
        return mysql_parser

    def execute(self, params, **kwargs):
        record_name = kwargs.pop('record', None)
        if not record_name:
            ls = RecordListCommand()
            return ls.execute(params, record_type=['databaseCredentials', 'serverCredentials'], verbose=True, **kwargs)

        record = find_record(params, record_name, types=['databaseCredentials', 'serverCredentials'])

        login = BaseConnectCommand.get_record_field(record, 'login')
        if not login:
            logging.warning('Record "%s" does not have login.', record.title)
            return

        host = BaseConnectCommand.get_record_field(record, 'host')
        if not host:
            logging.warning('Record "%s" does not have host.', record.title)
            return
        host_name, _, port = host.partition(':')

        self.run_at_the_end.clear()

        options = self.get_extra_options(params, record, 'mysql')
        self.command = f'mysql{options}'
        self.command += f' --host {host_name} --user {login}'
        if port:
            self.command += f' --port {port}'

        password = BaseConnectCommand.get_record_field(record, 'password')
        if password:
            os.putenv('MYSQL_PWD', password)

            def clear_env():
                os.putenv('MYSQL_PWD', '')
            self.run_at_the_end.append(clear_env)

        query = kwargs.get('query')
        if isinstance(query, list) and len(query) > 0:
            self.command += ' --execute \"' + ' '.join(query) + '\"'

        logging.info('Connecting to "%s" ...', record.title)
        self.execute_shell()


class ConnectPostgresCommand(BaseConnectCommand):
    def get_parser(self):
        return postgres_parser

    def execute(self, params, **kwargs):
        record_name = kwargs.pop('record', None)
        if not record_name:
            ls = RecordListCommand()
            return ls.execute(params, record_type=['databaseCredentials', 'serverCredentials'], verbose=True, **kwargs)

        record = find_record(params, record_name, types=['databaseCredentials', 'serverCredentials'])

        login = BaseConnectCommand.get_record_field(record, 'login')
        if not login:
            logging.warning('Record "%s" does not have user name.', record.title)
            return
        host = BaseConnectCommand.get_record_field(record, 'host')
        if not host:
            logging.warning('Record "%s" does not have host.', record.title)
            return
        host_name, _, port = host.partition(':')

        database = kwargs.get('database')
        if not database:
            database = BaseConnectCommand.get_custom_field(record, 'database')
        if not database:
            database = 'template1'
            logging.info(f'\nConnecting to the default database: {database}\n')

        self.command = f'{postgresql} {self.extra_parameters} -h {host_name}'
        if port:
            self.command += f' -p {port}'
        self.command += f' -U {login} -w {database}'
        self.run_at_the_end.clear()

        password = BaseConnectCommand.get_record_field(record, 'password')
        if password:
            os.putenv('PGPASSWORD', password)

            def clear_env():
                os.putenv('PGPASSWORD', '')
            self.run_at_the_end.append(clear_env)

        logging.info('Connecting to "%s" ...', record.title)
        self.execute_shell()


class ConnectRdpCommand(BaseConnectCommand):
    def get_parser(self):
        return rdp_parser

    def execute(self, params, **kwargs):
        record_name = kwargs.pop('record', None)
        if not record_name:
            ls = RecordListCommand()
            return ls.execute(params, record_type=['serverCredentials'], verbose=True, **kwargs)

        record = find_record(params, record_name, types=['serverCredentials'])

        login = BaseConnectCommand.get_record_field(record, 'login')
        if not login:
            logging.warning('Record "%s" does not have user name.', record.title)
            return

        host = BaseConnectCommand.get_record_field(record, 'host')
        if not host:
            logging.warning('Record "%s" does not have host.', record.title)
            return
        host_name, _, port = host.partition(':')

        password = BaseConnectCommand.get_record_field(record, 'password')
        if password:
            os.system(f'cmdkey /generic:{host_name} /user:{login} /pass:{password} > NUL')

            def clear_password():
                os.system(f'cmdkey /delete:{host_name} > NUL')
            self.run_at_the_end.append(clear_password)

        self.command = f'mstsc /v:{host_name}'
        if port:
            self.command += ':' + port

        logging.info('Connecting to "%s" ...', record.title)
        self.execute_shell()


connect_command_description = '''
Connect Command Syntax Description:
This command reads the custom fields for names starting with "connect:"
  connect:<name>                                    command 
  connect:<name>:description                        command description
  connect:<name>:ssh-key:<key-comment>              ssh private key to add to ssh-agent
  connect:<name>:env:<Environment Variable To Set>  sets environment variable
Connection command may contain template parameters.
Parameter syntax is ${<parameter_name>}
Supported parameters:
    ${user_email}                   Keeper user email address
    ${login}                        Record login
    ${password}                     Record password
    ${file:<attachment_name>}       stores attachment into temporary file. parameter is replaced with temp file name
    ${body:<attachment_name>}       content of the attachment file.
    ${<custom_field_name>}          custom field value
SSH Example:
Title: SSH to my Server via Gateway
Custom Field 1 Name: connect:my_server:description
Custom Field 1 Value: Production Server Inside Gateway
Custom Field 2 Name: connect:my_server
Custom Field 2 Value: ssh -o "ProxyCommand ssh -i ${file:gateway.pem} ec2-user@gateway.mycompany.com -W %h:%p" -i ${file:server.pem} ec2-user@server.company.com
File Attachments:
gateway.pem
server.pem
To initiate connection: "connect my_server"
Connect to Postgres Example:
Title:    Postgres
Login:    PGuser
Password: **************
Custom Field 1 Name:  connect:postgres
Custom Field 1 Value: psql --host=11.22.33.44 --port=3306 --username=${login} --dbname=postgres --no-password
Custom Field 2 Name:  connect:postgres:env:PGPASSWORD
Custom Field 2 Value: ${password}
To initiate connection: "connect postgres"
'''


endpoint_pattern = re.compile(r'^connect:([^:]+)$')
endpoint_desc_pattern = re.compile(r'^connect:([^:]+):description$')


class ConnectEndpoint:
    def __init__(self, name, description, record_uid, record_title, paths):
        self.name = name                    # type: str
        self.description = description      # type: str
        self.record_uid = record_uid        # type: str
        self.record_title = record_title    # type: str
        self.paths = paths                  # type: list


class ConnectCommand(BaseConnectCommand):
    LastRevision = 0        # int
    Endpoints = []          # type: [ConnectEndpoint]

    def get_parser(self):
        return connect_parser

    def execute(self, params, **kwargs):
        if kwargs.pop('syntax_help', None):
            logging.info(connect_command_description)
            return

        ConnectCommand.find_endpoints(params)

        endpoint = kwargs.pop('endpoint', None)
        if endpoint:
            endpoints = [x for x in ConnectCommand.Endpoints if x.name == endpoint]
            if not endpoints:
                rpos = endpoint.rfind(':')
                if rpos > 0:
                    try_path = endpoint[:rpos]
                    endpoint_name = endpoint[rpos+1:]
                else:
                    try_path = endpoint
                    endpoint_name = ''
                record_uid = ''
                if try_path in params.record_cache:
                    record_uid = try_path
                else:
                    rs = try_resolve_path(params, try_path)
                    if rs is not None:
                        folder, title = rs
                        if folder is not None and title is not None:
                            folder_uid = folder.uid or ''
                            if folder_uid in params.subfolder_record_cache:
                                for uid in params.subfolder_record_cache[folder_uid]:
                                    r = KeeperRecord.load(params, uid)
                                    if r:
                                        if r.title.lower() == title.lower():
                                            record_uid = uid
                                            break
                if record_uid:
                    endpoints = [x for x in ConnectCommand.Endpoints
                                 if x.record_uid == record_uid and endpoint_name in {'', x.name}]

            if len(endpoints) > 0:
                if len(endpoints) == 1:
                    record = KeeperRecord.load(params, endpoints[0].record_uid)
                    if record:
                        self.connect_endpoint(params, endpoints[0].name, record, **kwargs)
                else:
                    logging.warning("Connect endpoint '%s' is not unique", endpoint)
                    ConnectCommand.dump_endpoints(endpoints)
                    logging.info("Use full endpoint path: /<Folder>/<Title>[:<Endpoint>]")
                    folder = endpoints[0].paths[0] if len(endpoints[0].paths) > 0 else '/'
                    logging.info('Example: connect "%s/%s:%s"', folder, endpoints[0].record_title, endpoints[0].name)
            else:
                logging.info("Connect endpoint '%s' not found", endpoint)
        else:
            if ConnectCommand.Endpoints:
                return ConnectCommand.dump_endpoints(ConnectCommand.Endpoints, **kwargs)
            else:
                logging.info("No connect endpoints found")

    @staticmethod
    def dump_endpoints(endpoints, **kwargs):   # type: (List[ConnectEndpoint], ...) -> Optional[str]
        sort_by = kwargs.get('sort_by') or ''
        filter_by = kwargs.get('filter_by') or ''
        logging.info("Available connect endpoints")
        if filter_by:
            logging.info('Filtered by "%s"', filter_by)
            filter_by = filter_by.lower()
        fmt = kwargs.get('format', '')
        verbose = fmt != 'table'

        headers = ['endpoint', 'description', 'record_title', 'record_uid', 'folders']
        if fmt != 'json':
            headers = [field_to_title(x) for x in headers]

        table = []
        for endpoint in endpoints:
            title = endpoint.record_title
            folder = endpoint.paths[0] if len(endpoint.paths) > 0 else ''
            if filter_by:
                if not any([x for x in [endpoint.name.lower(), title.lower(), endpoint.record_uid, folder.lower()] if x.find(filter_by) >= 0]):
                    continue
            if not verbose and len(title) > 23:
                title = title[:20] + '...'
            table.append([endpoint.name, endpoint.description or '', title, endpoint.record_uid, folder])

        sorted_by = 4 if sort_by == 'folder' else 2 if sort_by == 'title' else 0
        return dump_report_data(table, headers, fmt=fmt, output=kwargs.get('output'), row_number=True, sort_by=sorted_by)

    @staticmethod
    def find_endpoints(params):   # type: (KeeperParams) -> None
        if ConnectCommand.LastRevision < params.revision:
            ConnectCommand.LastRevision = params.revision
            ConnectCommand.Endpoints.clear()
            for record_uid in params.record_cache:
                record = KeeperRecord.load(params, record_uid)
                if not record:
                    continue
                endpoints = []
                endpoints_desc = {}
                if isinstance(record, PasswordRecord):
                    for field in record.custom:
                        m = endpoint_pattern.match(field.name)
                        if m:
                            endpoints.append(m[1])
                        else:
                            m = endpoint_desc_pattern.match(field.name)
                            if m:
                                endpoints_desc[m[1]] = field.value or ''
                elif isinstance(record, TypedRecord):
                    for field in record.custom:
                        if field.type and field.type != 'text':
                            continue
                        m = endpoint_pattern.match(field.label)
                        if m:
                            endpoints.append(m[1])
                        else:
                            m = endpoint_desc_pattern.match(field.label)
                            if m:
                                endpoints_desc[m[1]] = field.get_default_value(str) or ''
                if endpoints:
                    paths = []
                    for folder_uid in find_folders(params, record_uid):
                        path = '/' + get_folder_path(params, folder_uid, '/')
                        paths.append(path)
                    for endpoint in endpoints:
                        epoint = ConnectEndpoint(endpoint, endpoints_desc.get(endpoint) or '', record_uid, record.title, paths)
                        ConnectCommand.Endpoints.append(epoint)
            ConnectCommand.Endpoints.sort(key=lambda x: x.name)

    @staticmethod
    def get_fields_by_patters(record, pattern):   # type: (KeeperRecord, str) -> Iterable[Tuple[str, str]]
        if isinstance(record, PasswordRecord):
            return ((x.name, x.value) for x in record.custom if x.name.lower().startswith(pattern))

        if isinstance(record, TypedRecord):
            return ((x.label, x.get_default_value()) for x in record.custom
                    if (x.type or 'text') == 'text' and (x.label or '').lower().startswith(pattern))

    @staticmethod
    def add_ssh_keys(params, endpoint, record, temp_files):
        # type: (KeeperParams, str, KeeperRecord, List[str]) -> Iterable[Callable]
        pk = try_extract_private_key(params, record)
        if pk:
            private_key, passphrase = pk
            to_delete = add_ssh_key(private_key, passphrase if passphrase else None, record.title)
            if to_delete:
                yield to_delete

        key_prefix = f'connect:{endpoint}:ssh-key'
        for cf_name, cf_value in ConnectCommand.get_fields_by_patters(record, key_prefix):
            key_name = cf_name[len(key_prefix)+1:] or 'Commander'
            parsed_values = []
            while True:
                m = endpoint_parameter_pattern.search(cf_value)
                if not m:
                    break
                p = m.group(1)
                val = ConnectCommand.get_parameter_value(params, record, p, temp_files)
                if not val:
                    raise Exception(f'Add ssh-key. Failed to resolve key parameter: {p}')
                parsed_values.append(val)
                cf_value = cf_value[m.end():]
            if len(parsed_values) > 0:
                cf_value = cf_value.strip()
                if cf_value:
                    parsed_values.append(cf_value)
                to_delete = add_ssh_key(parsed_values[0], parsed_values[1] if len(parsed_values) > 1 else None, key_name)
                if to_delete:
                    yield to_delete

    @staticmethod
    def add_environment_variables(params, endpoint, record, temp_files):
        # type: (KeeperParams, str, KeeperRecord, List[str]) -> Iterable[Callable]
        key_prefix = f'connect:{endpoint}:env:'
        for cf_name, cf_value in ConnectCommand.get_fields_by_patters(record, key_prefix):
            key_name = cf_name[len(key_prefix):]
            if not key_name:
                continue
            while True:
                m = endpoint_parameter_pattern.search(cf_value)
                if not m:
                    break
                p = m.group(1)
                val = ConnectCommand.get_parameter_value(params, record, p, temp_files)
                if not val:
                    raise Exception('Add environment variable. Failed to resolve key parameter: {0}'.format(p))
                cf_value = cf_value[:m.start()] + val + cf_value[m.end():]
            if cf_value:
                os.putenv(key_name, cf_value)

                def clear_env():
                    os.putenv(key_name, '')
                yield clear_env

    def connect_endpoint(self, params, endpoint, record, **kwargs):
        # type: (KeeperParams, str, KeeperRecord, ...) -> None
        temp_files = []
        try:
            command = BaseConnectCommand.get_custom_field(record, f'connect:{endpoint}:pre')
            if command:
                command = BaseConnectCommand.get_command_string(params, record, command, temp_files, endpoint=endpoint)
                if command:
                    os.system(command)

            command = ConnectCommand.get_custom_field(record, f'connect:{endpoint}')
            if command:
                self.command = ConnectCommand.get_command_string(params, record, command, temp_files, endpoint=endpoint)
                if self.command:
                    self.run_at_the_end.extend(
                        ConnectCommand.add_ssh_keys(params, endpoint, record, temp_files))
                    self.run_at_the_end.extend(
                        ConnectCommand.add_environment_variables(params, endpoint, record, temp_files))

                    parameters = kwargs.get('parameters')
                    if isinstance(parameters, list) and len(parameters) > 0:
                        self.command += ' ' + ' '.join(parameters)

                    logging.info('Connecting to "%s" ...', record.title)
                    self.execute_shell()

            command = BaseConnectCommand.get_custom_field(record, f'connect:{endpoint}:post')
            if command:
                command = ConnectCommand.get_command_string(params, record, command, temp_files, endpoint=endpoint)
                if command:
                    os.system(command)
        finally:
            for file in temp_files:
                os.remove(file)



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2021 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#
import json
import logging
import sys

import requests

from . import __version__, display
from .params import KeeperParams


def is_up_to_date_version(params):  # type: (KeeperParams) -> dict

    curr_git_version = None
    version_comparison = None
    release_download_url = 'https://github.com/Keeper-Security/Commander/releases'

    try:
        repo_owner = "Keeper-Security"
        repo_name = "Commander"
        repo_latest_release_api_url = f'https://api.github.com/repos/{repo_owner}/{repo_name}/releases/latest'

        rs = requests.get(repo_latest_release_api_url, proxies=params.rest_context.proxies, timeout=3)

        rs_dict = json.loads(rs.text)

        latest_tag_name = rs_dict.get('tag_name')
        publish_date = rs_dict.get('published_at')
        release_url = rs_dict.get('html_url')

        release_details = {
            'tag': latest_tag_name,
            'publish_date': publish_date,
            'release_url': release_url
        }
        release_download_url = release_details.get('release_url')
        this_app_version = __version__
        curr_git_version = release_details.get('tag')[1:]

        version_comparison = __version_compare(this_app_version, curr_git_version)
    except Exception as e:
        logging.debug(e)

    return {
        'is_up_to_date': version_comparison >= 0 if version_comparison else None,
        'current_github_version': curr_git_version,
        'new_version_download_url': release_download_url
    }


# Determine if this app is running as a python script or an installed binary (PyInstaller)
def is_binary_app():
    # see: https://pyinstaller.readthedocs.io/en/stable/runtime-information.html#run-time-information
    return getattr(sys, 'frozen', False) and hasattr(sys, '_MEIPASS')


def __version_compare(v1, v2):

    """
      Compare two Commander version versions and will return:
         1 if version 1 is bigger
         0 if equal
        -1 if version 2 is bigger
    """
    # This will split both the versions by '.'
    arr1 = v1.split(".")
    arr2 = v2.split(".")
    n = len(arr1)
    m = len(arr2)

    # converts to integer from string
    arr1 = [int(i) for i in arr1]
    arr2 = [int(i) for i in arr2]

    # compares which list is bigger and fills
    # smaller list with zero (for unequal delimeters)
    if n > m:
        for i in range(m, n):
            arr2.append(0)
    elif m > n:
        for i in range(n, m):
            arr1.append(0)

    # returns 1 if version 1 is bigger and -1 if
    # version 2 is bigger and 0 if equal
    for i in range(len(arr1)):
        if arr1[i] > arr2[i]:
            return 1
        elif arr2[i] > arr1[i]:
            return -1
    return 0


def welcome_print_version(params):

    this_app_version = __version__

    ver_info = is_up_to_date_version(params)

    if ver_info.get('is_up_to_date') is None:
        logging.debug(display.bcolors.WARNING + "It appears that the internet connection is offline." + display.bcolors.ENDC)

    elif not ver_info.get('is_up_to_date'):
        print(display.bcolors.WARNING +
              (" Your version of the Commander CLI is %s, the current version is %s.\n Use the ‘version’ "
               "command for more details.\n") % (this_app_version, ver_info.get('current_github_version')) + display.bcolors.ENDC
              )
    else:
        pass
        # print("Your version of the Commander CLI is %s." % this_app_version)
